/*
 * io-djgpp.c - DJGPP support routines for io test
 * Copyright (C) 2000 by Richard Dawe
 *
 * The io test uses pipe(). Unfortunately, DJGPP does not have a working
 * pipe() function. This is because pipes are usually between apps. However,
 * we need an intra-program pipe, which can be provided using DJGPP's File
 * System Extensions.
 */

/* TODO: Dynamically resize the pipe buffer. */

#ifdef __DJGPP__

#include <stddef.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/fsext.h>

/* This is a bit of a hack, but this is buffer is large enough for
 * mpz/tests/io.c to complete successfully. */
#define IP_NUMCHUNKS 1024
#define IP_CHUNKSIZE 1024
#define IP_BUFSIZE (IP_NUMCHUNKS * IP_CHUNKSIZE)

typedef struct _INTRA_PIPE {
  void   * buf;
  size_t   bufsize;
  int      fd_r;
  void   * pos_r;
  int      fd_w;
  void   * pos_w;
} INTRA_PIPE;

/* --------------------
 * - intra_pipe_fsext -
 * -------------------- */

/* This only supports reading, writing and closing - everything else will
 * fail. */

static int intra_pipe_fsext (__FSEXT_Fnumber n, int *rv, va_list args)
{
  int fd;
  INTRA_PIPE *ip = NULL;
  void *buf = NULL;
  size_t bufsize = 0;
  ptrdiff_t d = 0;
  ssize_t s = 0;
  
  switch(n)
    {
    case __FSEXT_close:
      fd = va_arg(args, int);
      ip = (INTRA_PIPE *) __FSEXT_get_data(fd);
      
      __FSEXT_set_data(fd, NULL);

      if (ip != NULL)
	{
	  if (ip->fd_r == fd)
	    {
	      _close(fd);
	      ip->fd_r  = -1;
	      ip->pos_r = NULL;
	    }

	  if (ip->fd_w == fd)
	    {
	      _close(fd);
	      ip->fd_w  = -1;
	      ip->pos_w = NULL;
	    }

	  /* Free data structures, now nothing refers to this fd pair. */
	  if ((ip->fd_w == -1) && (ip->fd_r == -1))
	    {
	      free(ip->buf);
	      free(ip);
	      ip = NULL;
	    }
	}
      
      *rv = 0;
      break;

    case __FSEXT_read:
      fd      = va_arg(args, int);
      buf     = va_arg(args, void *);
      bufsize = va_arg(args, size_t);
      ip      = (INTRA_PIPE *) __FSEXT_get_data(fd);

      if ((ip == NULL) || (ip->fd_r == -1))
	{
	  errno = EBADF;
	  *rv   = -1;
	  break;
	}

      if (fd != ip->fd_r)
        {
	  /* TODO: There should be a better code than this. */
	  errno = EACCES;
	  *rv   = -1;
	  break;
	}

      /* How much data is there to read? */
      d = ip->pos_w - ip->pos_r;
      s = (ssize_t) d;

      if (s == 0)
	{
	  /* No data => what shall we do? Technically this should block
	   * forever, but let's fail. */
	  errno = EBUSY;
	  *rv   = -1;
	  break;
	}
      
      /* If there's enough room, read all data. */
      if (s > bufsize)
	s = bufsize;

      memcpy(buf, ip->pos_r, s);
      ip->pos_r += s;
      *rv = s;      
      break;

    case __FSEXT_write:
      fd      = va_arg(args, int);
      buf     = va_arg(args, void *);
      bufsize = va_arg(args, size_t);
      ip      = (INTRA_PIPE *) __FSEXT_get_data(fd);

      if ((ip == NULL) || (ip->fd_w == -1))
	{
	  errno = EBADF;
	  *rv   = -1;
	  break;
	}

      if (fd != ip->fd_w)
        {
	  /* TODO: There should be a better code than this. */
	  errno = EACCES;
	  *rv   = -1;
	  break;
	}

      /* Who are you talking to? There's no-one there! */
      if (ip->fd_r == -1)
	{
	  errno = EPIPE;
	  *rv   = -1;
	  break;
	}
      
      /* How much space is left? */
      d = ip->pos_w - ip->buf;
      s = IP_BUFSIZE - (ssize_t) d;
      if (s <= 0)
	{
	  errno = ENOSPC;
	  *rv   = -1;
	  break;
	}

      /* If there's enough room, write all data. */
      if (s > bufsize)
	s = bufsize;

      memcpy(ip->pos_w, buf, s);
      ip->pos_w += s;
      
      *rv = s;
      break;
      
    default:
      /* Fail by default */
      *rv = -1;
      break;
    }

  /* Always emulate */
  return(1);
}

/* --------
 * - pipe -
 * -------- */

int pipe (int fildes[2])
{
  int fd[2] = { 0, 0 };
  INTRA_PIPE *ip = NULL;
  void *buf = NULL;
  size_t bufsize = IP_BUFSIZE;

  /* Allocate the file descriptors. */
  fd[0] = __FSEXT_alloc_fd(intra_pipe_fsext);
  if (fd[0] < 0)
    return(-1);
  
  fd[1] = __FSEXT_alloc_fd(intra_pipe_fsext);
  if (fd[1] < 0)
    {
      _close(fd[0]);
      return(-1);
    }

  /* Allocate some memory for the associated data structure. */
  ip = (INTRA_PIPE *) malloc(sizeof(*ip));
  if (ip == NULL)
    {
      _close(fd[0]);
      _close(fd[1]);
      errno = ENOMEM;
      return(-1);
    }

  buf = malloc(bufsize);
  if (buf == NULL)
    {
      free(ip);
      _close(fd[0]);
      _close(fd[1]);
      errno = ENOMEM;
      return(-1);
    }
  
  /* Set up the data structure. */
  bzero(ip, sizeof(*ip));
  
  ip->buf     = buf;
  ip->bufsize = bufsize;
  /* TODO: Is the first fd in a pair always read-only, and the second
   * write-only? */
  ip->fd_r    = fd[0];
  ip->pos_r   = buf;
  ip->fd_w    = fd[1];
  ip->pos_w   = buf;

  /* Now associate the data with the fds. */
  if (__FSEXT_set_data(fd[0], (void *) ip) == NULL)
    {
      free(ip->buf);
      free(ip);
      _close(fd[0]);
      _close(fd[1]);
      errno = ENOMEM;
      return(-1);
    }

  if (__FSEXT_set_data(fd[1], (void *) ip) == NULL)
    {
      free(ip->buf);
      free(ip);
      _close(fd[0]);
      _close(fd[1]);
      errno = ENOMEM;
      return(-1);
    }
  
  /* Done */
  fildes[0] = fd[0], fildes[1] = fd[1];
  return(0);
}

#endif /* __DJGPP__ */
